
from bottle import response
from .db import Visitor, User


class AuthCookie(object):

    cookie_key = 'asdf82jfooapAOOOOOaf'
    cookie_secret = 'nasdf423jndsfAJAfa'

    def __init__(self, user_id):
        self.user_id = user_id

    @property
    def user(self):
        return s.query(User).filter(User.id == self.user_id).first()


class VisitorPrefsCookie(object):

    cookie_key = 'jkasdfa99a9a9'
    cookie_secret = 'nakNJKNJANoi*8f**'

    def __init__(self, cookie_uuid):
        self.cookie_uuid = cookie_uuid

    @property
    def visitor_prefs(self):
        return (s.query(Visitor)
                .filter(Visitor.cookie_uuid == self.cookie_uuid).first())


def set_cookie(cookie):
    response.set_cookie(cookie.cookie_key, cookie, cookie.cookie_secret,
                        path='/')


def wipe_cookie(cookie):
    response.set_cookie(cookie.cookie_key, "")



"""
xychan - a simple Python message board system
Copyright (C) 2010 Mike Verdone

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Running xychan
--------------

::

    from xychan import app
    app.configure_db("sqlite:///database.db")
    app.configure_image_dir("path/to/image_dir")

"app" is the xychan WSGI app which you can run through pretty much any
WSGI server. The simplest is CGI::

    from wsgiref.handlers import CGIHandler
    CGIHandler().run(app)

Check out the file ``xychan.dh.cgi`` for a working CGI script, and
``xychan.dh.passenger`` for a Passenger WSGI deployment.

"""

from bottle import TEMPLATE_PATH, default_app, request, cookie_is_encoded
import os

STATIC_PATH = __path__[0] + os.sep + 'static'

TEMPLATE_PATH.insert(0, __path__[0] + os.sep + 'templates')

from basics import *
from .db import DbSessionMiddleware, configure_db
from .image_store import configure_image_dir
from .context import context_middleware

app = context_middleware(
    DbSessionMiddleware(
        default_app()))
app.configure_image_dir = configure_image_dir
app.configure_db = configure_db

__all__ = ['app']




from util import *

@post("/mod/trash_post/:post_id#[0-9]+#", name="trash_post")
@view("message.tpl")
@admin_only
def trash_post(post_id):
    post = s.query(Post).filter(Post.id == post_id).first()
    if post:
        redirect = url("board", board_name=post.thread.board.short_name)
        if post.is_first:
            thread = post.thread
            [s.delete(p) for p in thread.posts]
            s.delete(thread)
        else:
            s.delete(post)
    else:
        redirect = url("index")
    return dict(message="Trashed", redirect=redirect)




import re
import cgi
from uuid import uuid4
from mimetypes import guess_type

from util import *

_link_line_re = re.compile(r"&gt;&gt; *([0-9]+) *")

def get_board_or_die(s, board_name):
    board = s.query(Board).filter(Board.short_name == board_name).first()
    if not board:
        raise HTTPError(404, "No such board.")
    return board


def get_thread_in_board_or_die(s, board, thread_id):
    thread = (s.query(Thread)
              .filter(Thread.id == thread_id)
              .filter(Thread.board == board)
              .first())
    if not thread:
        raise HTTPError(404, "No such thread.")
    return thread


def sanitize_content(s, content):
    lines = []
    for line in content.split('\n'):
        line = cgi.escape(line)
        m = _link_line_re.match(line)
        if m:
            post = s.query(Post).filter(Post.id == m.groups()[0]).first()
            if post:
                line = (u"<a href=\""
                        + url('thread', board_name=post.thread.board.short_name,
                              thread_id=post.thread.id)
                        + "#" + str(post.id) + "\">"
                        + line
                        + "</a>")
        lines.append(line)
    return '<br>\n'.join(lines)


@get('/', name='index')
@view('index.tpl')
def index():
    boards = s.query(Board).order_by(Board.short_name).all()
    return dict(boards=boards)


@get('/setup')
def create_a_board():
    b = Board(short_name='test')
    s.add(b)
    t = Thread(board=b)
    s.add(t)
    s.add(Post(
            thread=t, content="This is a post", poster_ip='1.2.3.4',
            poster_name="poster_name", subject="subject"))
    u = User(username='admin')
    u.password = 'adminadmin1'
    s.add(u)
    return dict()


from login import *

from moderation import *


@get('/t_/:image', name='thumb')
@cache_forever
def get_thumbnail(image):
    thumb_data = fetch_thumb(image)
    response.content_type = 'image/' + image.split('.')[-1]
    return thumb_data


@get('/i_/:image', name='image')
@cache_forever
def get_image(image):
    image_data = fetch_image(image)
    response.content_type = 'image/' + image.split('.')[-1]
    return image_data


@get('/:board_name/', name='board')
@get('/:board_name/page/:page#[0-9]+#', name='board_page')
@view('board.tpl')
def board(board_name, page=None):
    board = get_board_or_die(s, board_name)
    page_offset = int(page or 0) * c.threads_per_page
    threads = (s.query(Thread)
               .filter(Thread.board == board)
               .order_by(desc(Thread.last_post_time))
               .offset(page_offset).limit(c.threads_per_page))
    threads = [thread for thread in threads if thread.posts]
    return dict(board=board, threads=threads)


def remember_poster_prefs(poster_name, use_tripcode):
    vp = c.visitor_prefs
    if not vp:
        vp = Visitor(cookie_uuid=str(uuid4()))
        s.add(vp)
        set_cookie(VisitorPrefsCookie(vp.cookie_uuid))
    vp.poster_name = poster_name
    vp.use_tripcode = use_tripcode


def handle_post(board_name, thread_id=None):
    board = get_board_or_die(s, board_name)
    if thread_id:
        thread = get_thread_in_board_or_die(s, board, thread_id)
    else:
        thread = Thread(board=board, last_post_time=func.now())
        s.add(thread)
    image_key = None
    img = request.files.get('image')
    if img is not None: # LOL WAT
        image_key = store_image(img.value)
    if not (image_key or get_uni('content').strip()):
        return dict(
            message="No, you must provide an image or some text in your post",
            redirect=url('board', board_name=board.short_name))
    poster_name = get_uni('poster_name')
    use_tripcode = bool(get_uni('use_tripcode'))
    remember_poster_prefs(poster_name, use_tripcode)
    s.add(Post(thread=thread,
               content=sanitize_content(s, get_uni('content')),
               poster_name=poster_name,
               subject=get_uni('subject'),
               poster_ip=request.get('REMOTE_ADDR', '0.0.0.0'),
               image_key=image_key,
               visitor_id=(c.visitor_prefs.id
                           if use_tripcode else None)))
    return dict(message="Post successful",
                redirect=url('board', board_name=board.short_name))


@post('/:board_name/post', name="post_thread")
@view('message.tpl')
def post_thread(board_name):
    return handle_post(board_name)


@get('/:board_name/:thread_id#[0-9]+#/', name='thread')
@get('/:board_name/:thread_id#[0-9]+#')
@view('thread.tpl')
def thread(board_name, thread_id):
    board = get_board_or_die(s, board_name)
    thread = get_thread_in_board_or_die(s, board, thread_id)
    return dict(board=board, thread=thread)


@post('/:board_name/:thread_id#[0-9]+#/post', name="post_reply")
@view('message.tpl')
def post_reply(board_name, thread_id):
    return handle_post(board_name, thread_id)


get(r'/:board_name#[^.]+#')(board)


@get(r'/:file#.+\..+#', name='static')
@cache_forever
def static(file):
    if file.split('.')[-1] in ('py', 'pyo', 'pyc'):
        raise HTTPError(404, "Not found")
    mime_type, encoding = guess_type(file)
    if mime_type:
        response.content_type = mime_type
    return open(STATIC_PATH + os.sep + file, 'rb').read()


@error(404)
def error404(msg):
    return "Page not found."




from __future__ import with_statement

import math
from sha import sha

from sqlalchemy import (
    create_engine,
    Table, Column, Integer, String, MetaData, ForeignKey, DateTime,
    func, Boolean,
    )
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, backref
from sqlalchemy.sql.expression import desc

from base62 import num_encode


metadata = MetaData()
Session = sessionmaker()
Base = declarative_base(metadata=metadata)

def random_str():
    import random
    return num_encode(random.randint(0, 18446744073709551615L))


class SessionContextMgr(object):
    def __init__(self):
        self.lvl = 0

    def __enter__(self):
        if not self.lvl:
            self.session = Session()
            __builtins__['s'] = self.session
        self.lvl += 1
        return self.session

    def __exit__(self, typ, value, tb):
        self.lvl -= 1
        if not self.lvl:
            if typ:
                self.session.rollback()
            else:
                self.session.commit()
            self.session = None
            del __builtins__['s']

active_session = SessionContextMgr()


def DbSessionMiddleware(wrapped_app):
    def _DbSessionMiddleware(environ, start_response):
        with active_session as s:
            environ['db_session'] = s
            return wrapped_app(environ, start_response)
    return _DbSessionMiddleware


class Board(Base):
    __tablename__ = 'boards'

    id = Column(Integer, primary_key=True)
    short_name = Column(String, nullable=False, unique=True)

    @property
    def num_pages(self):
        return int(math.ceil(float(len(self.threads)) / c.threads_per_page))


class Thread(Base):
    __tablename__ = 'threads'

    id = Column(Integer, primary_key=True)
    board_id = Column(Integer, ForeignKey('boards.id'), nullable=False)
    last_post_time = Column(DateTime)

    board = relationship(Board,
                         backref=backref(
            'threads', order_by=desc(last_post_time)))


    @property
    def short_view_posts(self):
        posts = self.posts
        if len(posts) < 5:
            return posts[:]
        else:
            return posts[0:1] + posts[-3:]


class Post(Base):
    __tablename__ = 'posts'

    id = Column(Integer, primary_key=True)
    thread_id = Column(Integer, ForeignKey('threads.id'), nullable=False)
    posted = Column(DateTime, nullable=False, default=func.now())
    poster_ip = Column(String, nullable=False)
    poster_name = Column(String, nullable=False)
    subject = Column(String, nullable=False)
    content = Column(String, nullable=False)
    image_key = Column(String)
    visitor_id = Column(Integer, ForeignKey('visitors.id'),
                        nullable=True)

    thread = relationship(Thread, backref=backref('posts', order_by=id))
    visitor = relationship('Visitor', backref=backref('posts', order_by=id))

    @property
    def is_first(self):
        return self == self.thread.posts[0]

    @property
    def tripcode(self):
        if self.visitor_id:
            return self.visitor.tripcode

    @property
    def formatted_date(self):
        return self.posted.strftime('%d %B %Y %H:%M')


class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    username = Column(String, nullable=False, unique=True)
    password_hash = Column(String, nullable=False)

    def _calc_password(self, password):
        sig = sha()
        sig.update(str(self.username))
        sig.update('saltysalt8828282')
        sig.update(password)
        return sig.hexdigest()


    def _get_password(self):
        raise Exception("Write only")
    def _set_password(self, v):
        self.password_hash = self._calc_password(v)
        return locals()
    password = property(_get_password, _set_password)


    def verify_password(self, password):
        return self.password_hash == self._calc_password(password)


class Visitor(Base):
    __tablename__ = 'visitors'

    id = Column(Integer, primary_key=True)
    cookie_uuid = Column(String, unique=True, nullable=False)
    poster_name = Column(String, nullable=False, default='')
    tripcode = Column(String, nullable=False, default=random_str)
    use_tripcode = Column(Boolean, nullable=False, default=False)


def configure_db(db_uri, echo=False):
    engine = create_engine(db_uri, echo=echo)
    metadata.bind = engine
    Session.bind = engine
    metadata.create_all()




# This came from:
# http://stackoverflow.com/questions/1119722/base-62-conversion-in-python

import string
ALPHABET = string.ascii_uppercase + string.ascii_lowercase + \
           string.digits + '-_'
ALPHABET_REVERSE = dict((c, i) for (i, c) in enumerate(ALPHABET))
BASE = len(ALPHABET)
SIGN_CHARACTER = '$'


def num_encode(n):
    if n < 0:
        return SIGN_CHARACTER + num_encode(-n)
    s = []
    while True:
        n, r = divmod(n, BASE)
        s.append(ALPHABET[r])
        if n == 0: break
    return ''.join(reversed(s))


def num_decode(s):
    if s[0] == SIGN_CHARACTER:
        return -num_decode(s[1:])
    n = 0
    for c in s:
        n = n * BASE + ALPHABET_REVERSE[c]
    return n

__all__ = ['num_encode', 'num_decode']




import os
try:
    from subprocess import check_output
except ImportError:
    from commands import getstatusoutput
    def check_output(cmdlist):
        cmd = ' '.join(cmdlist)
        status, output = getstatusoutput(cmd)
        if status:
            raise Exception("Command [%s] failed." % cmd)

from sha import sha

from .util import num_encode


def configure_image_dir(d):
    global IMAGE_DIR, THUMBS_DIR
    IMAGE_DIR = d
    THUMBS_DIR = d + os.sep + "thumbs"
configure_image_dir("_images")


def _im_identify_type(fn):
    out = check_output(["identify", "-format", "%m\n", fn]).split('\n')[0]
    return out.strip().lower()


def _im_make_thumb(fn):
    typ = fn[fn.rindex('.') + 1:]
    if typ in ('jpg', 'jpeg'):
        check_output(["convert", fn, "-resize", "150x150", "-quality", "60",
                      THUMBS_DIR + fn[fn.rindex(os.sep):fn.rindex('.')]
                      + '.jpeg'])
    elif typ in ('gif',):
        check_output(["convert", fn + r"[0]", "-resize", "150x150",
                      THUMBS_DIR + fn[fn.rindex(os.sep):fn.rindex('.')]
                      + '.gif'])
    elif typ in ('png',):
        check_output(["convert", fn, "-resize", "150x150",
                      THUMBS_DIR + fn[fn.rindex(os.sep):fn.rindex('.')]
                      + '.png'])
    else:
        raise Exception("Unknown image type")


def store_image(image_data):
    sig = sha()
    sig.update(image_data)
    key = num_encode(int(sig.hexdigest(), 16))
    fn = IMAGE_DIR + os.sep + key
    open(fn, 'w').write(image_data)
    typ = _im_identify_type(fn)
    real_fn = fn + '.' + typ
    check_output(["mv", fn, real_fn])
    _im_make_thumb(real_fn)
    return key + '.' + typ


def fetch_image(fn):
    if fn.count(os.sep):
        raise Exception("Uh no")
    return open(IMAGE_DIR + os.sep + fn).read()


def fetch_thumb(fn):
    if fn.count(os.sep):
        raise Exception("Uh no")
    return open(THUMBS_DIR + os.sep + fn).read()


__all__ = ['fetch_image', 'fetch_thumb', 'store_image']




import os

from bottle import (
    route, error, HTTPError, get, post, request, response, view, url,
    cookie_is_encoded,
    )

from base62 import *


def cache_forever(func):
    def _cache_forever(*args, **kwargs):
        response.headers['Expires'] = 'Sun, 17-Jan-2038 19:14:07 GMT'
        response.headers['Cache-control'] = 'public'
        return func(*args, **kwargs)
    return _cache_forever


def admin_only(func):
    def _admin_only(*args, **kwargs):
        if not c.user:
            raise HTTPError(403, 'Forbidden')
        return func(*args, **kwargs)
    return _admin_only


def get_uni(key):
    return request.forms.get(key, '').decode('utf8')


from image_store import *
from db import *
from cookies import *

from . import STATIC_PATH

__builtins__['url'] = url




from bottle import cookie_is_encoded, request
from cookies import *


class Context(object):

    def _fetch_cookie(self, cookie_def):
        cookie = None
        if cookie_is_encoded(request.COOKIES.get(cookie_def.cookie_key, '')):
            cookie = request.get_cookie(
                cookie_def.cookie_key,
                cookie_def.cookie_secret)
        return cookie

    @property
    def user(self):
        auth_cookie = self._fetch_cookie(AuthCookie)
        if auth_cookie:
            return auth_cookie.user
        else:
            return None

    @property
    def visitor_prefs(self):
        cookie = self._fetch_cookie(VisitorPrefsCookie)
        if cookie:
            return cookie.visitor_prefs

    threads_per_page = 10


def context_middleware(wrapped_app):
    def _context_middleware(environ, start_response):
        __builtins__['c'] = Context()
        try:
            return wrapped_app(environ, start_response)
        finally:
            del __builtins__['c']
    return _context_middleware



from setuptools import setup, find_packages
import sys, os

version = '0.2'

setup(name='xychan',
      version=version,
      description="A 'chan-style' web message board similar to Wakaba",
      long_description=open('README.md', 'r').read(),
      classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Communications :: BBS",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='4chan, chan, message board, Wakaba',
      author='Mike Verdone',
      author_email='mike.verdone@gmail.com',
      url='http://mike.verdone.ca/xychan',
      license='GNU General Public License v3 (GPL)',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          'bottle>=0.8.3',
          'sqlalchemy>=0.6',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )




from util import *

@get('/login', name='login')
@view('login.tpl')
def login():
    return dict()


@post('/log_me_in_please', name='login_submit')
@view('message.tpl')
def login_submit():
    user = (s.query(User).filter(User.username == request.POST.get('username'))
            .first())
    if (not user) or (not user.verify_password(request.POST.get('password'))):
        return dict(message="Nope, that's not it", redirect=url('login'))
    else:
        set_cookie(AuthCookie(user.id))
        return dict(message="You are now logged in", redirect=url('index'))


@post('/log_me_out_please', name='logout_submit')
@view('message.tpl')
def logout_submit():
    wipe_cookie(AuthCookie)
    return dict(message="Seeya later", redirect=url('index'))


